#if !defined(WIN32_LEAN_AND_MEAN)
#define WIN32_LEAN_AND_MEAN
#endif

#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <cmath>
#include <time.h>
#include "main.h"
bool drawWorld;

// Constants

#define APP_TITLE "OpenGL Procedural Terrain"

// GL_EXT_texture_filter_anisotropic
#define GL_TEXTURE_MAX_ANISOTROPY_EXT     0x84FE
#define GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT 0x84FF

// Function pointers for the ARB multitexturing functions
PFNGLMULTITEXCOORD2FARBPROC		glMultiTexCoord2fARB	 = NULL;
PFNGLACTIVETEXTUREARBPROC		glActiveTextureARB		 = NULL;

int				g_TextureSize = 512;//256; //(reflection/refraction/depth)
float			g_WaterHeight = 58.0f; //height above 0
float			g_WaterUV = 15.0f; //water tex scale
float			g_WaterFlow = 0.0010f; //flow speed
const float		kCausticScale = 4.0f;

const int       TERRAIN_REGIONS_COUNT = 4;

const float     HEIGHTMAP_ROUGHNESS = 1.2f;
const float     HEIGHTMAP_SCALE = 1.0;
const float     HEIGHTMAP_TILING_FACTOR = 12.0f;
const int       HEIGHTMAP_SIZE = 64;
const int       HEIGHTMAP_GRID_SPACING = 16;

const float     CAMERA_FOVX = 90.0f;
const float     CAMERA_ZFAR = HEIGHTMAP_SIZE * HEIGHTMAP_GRID_SPACING * 5.0f;
const float     CAMERA_ZNEAR = 1.0f;
const float     CAMERA_Y_OFFSET = 15.0f;
const Vector3   CAMERA_ACCELERATION(100.0f, 100.0f, 100.0f);
const Vector3   CAMERA_VELOCITY( 60.0f, 60.0f, 60.0f);
const Vector3   CAMERA_GOD_ACCELERATION(120.0f, 120.0f, 120.0f);
const Vector3   CAMERA_GOD_VELOCITY(90.0f, 90.0f, 90.0f);
const float		MAX_GOD_HEIGHT = 780; // maximum height of god mode cam 
int				ITER = 12; // starts with 5 iterations for Fault line algorithm (pretty water!!)
int				MAX_ITER = 500; //maximum iterations allowed for Fault line
float			BLUR = 0.23; // much much to blur the terrain (FIR filter) values from 0-1;

// Types

struct TerrainRegion
{
    float min;
    float max;
    GLuint texture;
    std::string filename;
};


// Globals

HWND                g_hWnd;
HDC                 g_hDC;
HGLRC               g_hRC;
HINSTANCE           g_hInstance;
int                 g_framesPerSecond;
int                 g_windowWidth;
int                 g_windowHeight;
int                 g_maxAnisotrophy;
bool                g_isFullScreen;
bool                g_hasFocus;
bool                g_enableVerticalSync;
bool                g_displayHelp;
bool                g_disableColorMaps;
bool				g_GodMode;
bool				faultline;
bool				wireframe;
float               g_lightDir[4] = {0.0f, 1.0f, 0.0f, 0.0f};
GLuint              g_nullTexture;
GLuint              g_terrainShader;
GLFont              g_font;
Terrain             g_terrain;
Camera              g_camera;
Vector3             g_cameraBoundsMax;
Vector3             g_cameraBoundsMin;
//global shader object 
CShader				g_Shader;
UINT				g_Texture[MAX_TEXTURES] = {0};

TerrainRegion g_regions[TERRAIN_REGIONS_COUNT] =
{
    // Terrain regions
    0.0f, 60.0f * HEIGHTMAP_SCALE, 0, "Data/textures/dirtb.jpg",
    61.0f * HEIGHTMAP_SCALE, 101.0f * HEIGHTMAP_SCALE, 0, "Data/textures/grassa.jpg",
    102.0f * HEIGHTMAP_SCALE, 203.0f * HEIGHTMAP_SCALE, 0, "Data/textures/rockc.jpg",
    204.0f * HEIGHTMAP_SCALE, 255.0f * HEIGHTMAP_SCALE, 0, "Data/textures/snowa.jpg"
};


// Functions Prototypes
	
bool	CreateTexture(UINT &texture, LPSTR strFileName);
void	CreateSkyBox(float x, float y, float z, float width, float height, float length);
void	RenderWorld(bool bRenderCaustics, float delta);
void    BindTexture(GLuint texture, GLuint unit);
void    Cleanup();
void    CleanupApp();
GLuint  CompileShader(GLenum type, const GLchar *pszSource, GLint length);
HWND    CreateAppWindow(const WNDCLASSEX &wcl, const char *pszTitle);
GLuint  CreateNullTexture(int width, int height);
void    EnableVerticalSync(bool enableVerticalSync);
void    GenerateTerrain(float tick);
float   GetElapsedTimeInSeconds();
Vector3 GetMovementDirection();
bool    Init();
void    InitApp();
void    InitGL();
GLuint  LinkShaders(GLuint vertShader, GLuint fragShader);
GLuint  LoadShaderProgram(const char *pszFilename, std::string &infoLog);
GLuint  LoadTexture(const char *pszFilename);
GLuint  LoadTexture(const char *pszFilename, GLint magFilter, GLint minFilter,
                    GLint wrapS, GLint wrapT);
void    PerformCameraCollisionDetection();
void    ProcessUserInput(float tick);
void    ReadTextFile(const char *pszFilename, std::string &buffer);
void    RenderFrame(float delta);
void    RenderTerrain(float delta);
void    RenderText();
void    SetProcessorAffinity();
void    ToggleFullScreen();
void    UpdateCamera(float elapsedTimeSec);
void    UpdateFrame(float elapsedTimeSec, float tick);
void    UpdateFrameRate(float elapsedTimeSec);
void    UpdateTerrainShaderParameters();
LRESULT CALLBACK WindowProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);


bool returnGod()
{ 
	return g_GodMode; 
}

bool returnFault() //faultline or DSF selected
{
	return faultline;
}

bool returnVsync()
{
	return g_enableVerticalSync;
}


// Functions

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
#if defined _DEBUG
    _CrtSetDbgFlag(_CRTDBG_LEAK_CHECK_DF | _CRTDBG_ALLOC_MEM_DF);
    _CrtSetReportMode(_CRT_ASSERT, _CRTDBG_MODE_FILE);
    _CrtSetReportFile(_CRT_ASSERT, _CRTDBG_FILE_STDERR);
#endif

    MSG msg = {0};
    WNDCLASSEX wcl = {0};

    wcl.cbSize = sizeof(wcl);
    wcl.style = CS_OWNDC | CS_HREDRAW | CS_VREDRAW;
    wcl.lpfnWndProc = WindowProc;
    wcl.cbClsExtra = 0;
    wcl.cbWndExtra = 0;
    wcl.hInstance = g_hInstance = hInstance;
    wcl.hIcon = LoadIcon(0, IDI_APPLICATION);
    wcl.hCursor = LoadCursor(0, IDC_ARROW);
    wcl.hbrBackground = 0;
    wcl.lpszMenuName = 0;
    wcl.lpszClassName = "GL2WindowClass";
    wcl.hIconSm = 0;

    if (!RegisterClassEx(&wcl))
        return 0;

    g_hWnd = CreateAppWindow(wcl, APP_TITLE);

    if (g_hWnd)
    {
        SetProcessorAffinity();

        if (Init())
        {
            ShowWindow(g_hWnd, nShowCmd);
            UpdateWindow(g_hWnd);

			float tick = 0;
			GetElapsedTimeInSeconds();
            while (true)
            {
				float delta = GetElapsedTimeInSeconds();
				tick += delta;
                while (PeekMessage(&msg, 0, 0, 0, PM_REMOVE))
                {
                    if (msg.message == WM_QUIT)
                        break;

                    TranslateMessage(&msg);
                    DispatchMessage(&msg);
                }

                if (msg.message == WM_QUIT)
                    break;

                if (g_hasFocus)
                {
                    UpdateFrame(delta, tick);
                    RenderFrame(tick);
                    SwapBuffers(g_hDC);
                }
                else
                {
                    WaitMessage();
                }
            }
        }

        Cleanup();
        UnregisterClass(wcl.lpszClassName, hInstance);
    }

    return static_cast<int>(msg.wParam);
}

LRESULT CALLBACK WindowProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch (msg)
    {
    case WM_ACTIVATE:
        switch (wParam)
        {
        default:
            break;

        case WA_ACTIVE:
        case WA_CLICKACTIVE:
            Mouse::instance().attach(hWnd);
            g_hasFocus = true;
            break;

        case WA_INACTIVE:
            if (g_isFullScreen)
                ShowWindow(hWnd, SW_MINIMIZE);
            Mouse::instance().detach();
            g_hasFocus = false;
            break;
        }
        break;
    
    case WM_DESTROY:
        PostQuitMessage(0);
        return 0;

    case WM_SIZE:
        g_windowWidth = static_cast<int>(LOWORD(lParam));
        g_windowHeight = static_cast<int>(HIWORD(lParam));
        break;

    default:
        Mouse::instance().handleMsg(hWnd, msg, wParam, lParam);
        Keyboard::instance().handleMsg(hWnd, msg, wParam, lParam);
        break;
    }

    return DefWindowProc(hWnd, msg, wParam, lParam);
}

void BindTexture(GLuint texture, GLuint unit)
{
    glActiveTexture(GL_TEXTURE0 + unit);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, texture);
}

void Cleanup()
{
    CleanupApp();

    if (g_hDC)
    {
        if (g_hRC)
        {
            wglMakeCurrent(g_hDC, 0);
            wglDeleteContext(g_hRC);
            g_hRC = 0;
        }

        ReleaseDC(g_hWnd, g_hDC);
        g_hDC = 0;
    }
}

void CleanupApp()
{
    for (int i = 0; i < TERRAIN_REGIONS_COUNT; ++i)
    {
        if (g_regions[i].texture)
        {
            glDeleteTextures(1, &g_regions[i].texture);
            g_regions[i].texture = 0;
        }
    }

    if (g_nullTexture)
    {
        glDeleteTextures(1, &g_nullTexture);
        g_nullTexture = 0;
    }

    if (g_terrainShader)
    {
        glUseProgram(0);
        glDeleteProgram(g_terrainShader);
        g_terrainShader = 0;
    }

    g_terrain.destroy();
    g_font.destroy();
}

GLuint CompileShader(GLenum type, const GLchar *pszSource, GLint length)
{
    GLuint shader = glCreateShader(type);

    if (shader)
    {
        GLint compiled = 0;

        glShaderSource(shader, 1, &pszSource, &length);
        glCompileShader(shader);
        glGetShaderiv(shader, GL_COMPILE_STATUS, &compiled);

        if (!compiled)
        {
            GLsizei infoLogSize = 0;
            std::string infoLog;

            glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &infoLogSize);
            infoLog.resize(infoLogSize);
            glGetShaderInfoLog(shader, infoLogSize, &infoLogSize, &infoLog[0]);

            throw infoLog;
        }
    }

    return shader;
}

HWND CreateAppWindow(const WNDCLASSEX &wcl, const char *pszTitle)
{
    DWORD wndExStyle = WS_EX_OVERLAPPEDWINDOW;
    DWORD wndStyle = WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU |
        WS_MINIMIZEBOX | WS_CLIPCHILDREN | WS_CLIPSIBLINGS;

    HWND hWnd = CreateWindowEx(wndExStyle, wcl.lpszClassName, pszTitle,
        wndStyle, 0, 0, 0, 0, 0, 0, wcl.hInstance, 0);

    if (hWnd)
    {
        int screenWidth = GetSystemMetrics(SM_CXSCREEN);
        int screenHeight = GetSystemMetrics(SM_CYSCREEN);
        int halfScreenWidth = screenWidth / 2;
        int halfScreenHeight = screenHeight / 2;
        int left = (screenWidth - halfScreenWidth) / 2;
        int top = (screenHeight - halfScreenHeight) / 2;
        RECT rc = {0};

        SetRect(&rc, left, top, left + halfScreenWidth, top + halfScreenHeight);
        AdjustWindowRectEx(&rc, wndStyle, FALSE, wndExStyle);
        MoveWindow(hWnd, rc.left, rc.top, rc.right - rc.left, rc.bottom - rc.top, TRUE);

        GetClientRect(hWnd, &rc);
        g_windowWidth = rc.right - rc.left;
        g_windowHeight = rc.bottom - rc.top;
    }

    return hWnd;
}

GLuint CreateNullTexture(int width, int height)
{
	// to turn off terrain textures
    int pitch = ((width * 32 + 31) & ~31) >> 3; // align to 4-byte boundaries
    std::vector<GLubyte> pixels(pitch * height, 255);
    GLuint texture = 0;
    glGenTextures(1, &texture);
    glBindTexture(GL_TEXTURE_2D, texture);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, width, height, 0, GL_BGRA,
        GL_UNSIGNED_BYTE, &pixels[0]);
    glBindTexture(GL_TEXTURE_2D, 0);
    return texture;
}

void EnableVerticalSync(bool enableVerticalSync)
{
    typedef BOOL (WINAPI * PFNWGLSWAPINTERVALEXTPROC)(GLint);
    static PFNWGLSWAPINTERVALEXTPROC wglSwapIntervalEXT =
        reinterpret_cast<PFNWGLSWAPINTERVALEXTPROC>(
        wglGetProcAddress("wglSwapIntervalEXT"));

    if (wglSwapIntervalEXT)
    {
        wglSwapIntervalEXT(enableVerticalSync ? 1 : 0);
        g_enableVerticalSync = enableVerticalSync;
    }
}

void GenerateTerrain(float tick)
{
	if(!faultline)
	{
		if (!g_terrain.generateUsingDiamondSquareFractal(HEIGHTMAP_ROUGHNESS, tick))
		   throw std::runtime_error("Failed to generate terrain.");
	}
	else
	{
		if (!g_terrain.generateUsingFaultline(ITER, BLUR, tick))
        throw std::runtime_error("Failed to generate terrain.");
	}
}

float GetElapsedTimeInSeconds()
{
	// high accurary timer that returns elapsed time in seconds
    static const int MAX_SAMPLE_COUNT = 50;
    static float frameTimes[MAX_SAMPLE_COUNT];
    static float timeScale = 0.0f;
    static float actualElapsedTimeSec = 0.0f;
    static INT64 freq = 0;
    static INT64 lastTime = 0;
    static int sampleCount = 0;
    static bool initialized = false;

    INT64 time = 0;
    float elapsedTimeSec = 0.0f;

    if (!initialized)
    {
        initialized = true;
        QueryPerformanceFrequency(reinterpret_cast<LARGE_INTEGER*>(&freq));
        QueryPerformanceCounter(reinterpret_cast<LARGE_INTEGER*>(&lastTime));
        timeScale = 1.0f / freq;
    }

    QueryPerformanceCounter(reinterpret_cast<LARGE_INTEGER*>(&time));
    elapsedTimeSec = (time - lastTime) * timeScale;
    lastTime = time;

    if (fabsf(elapsedTimeSec - actualElapsedTimeSec) < 1.0f)
    {
        memmove(&frameTimes[1], frameTimes, sizeof(frameTimes) - sizeof(frameTimes[0]));
        frameTimes[0] = elapsedTimeSec;

        if (sampleCount < MAX_SAMPLE_COUNT)
            ++sampleCount;
    }

    actualElapsedTimeSec = 0.0f;

    for (int i = 0; i < sampleCount; ++i)
        actualElapsedTimeSec += frameTimes[i];

    if (sampleCount > 0)
        actualElapsedTimeSec /= sampleCount;

    return actualElapsedTimeSec;
}

Vector3 GetMovementDirection()
{
    static bool moveForwardsPressed = false;
    static bool moveBackwardsPressed = false;
    static bool moveRightPressed = false;
    static bool moveLeftPressed = false;
    static bool moveUpPressed = false;
    static bool moveDownPressed = false;

    Vector3 direction(0.0f, 0.0f, 0.0f);
    Vector3 velocity = g_camera.getCurrentVelocity();
    Keyboard &keyboard = Keyboard::instance();
	// Movement
    if (keyboard.keyDown(Keyboard::KEY_UP) || keyboard.keyDown(Keyboard::KEY_W))
    {
        if (!moveForwardsPressed)
        {
            moveForwardsPressed = true;
            g_camera.setCurrentVelocity(velocity.x, velocity.y, 0.0f);
        }

        direction.z += 1.0f;
    }
    else
    {
        moveForwardsPressed = false;
    }

    if (keyboard.keyDown(Keyboard::KEY_DOWN) || keyboard.keyDown(Keyboard::KEY_S))
    {
        if (!moveBackwardsPressed)
        {
            moveBackwardsPressed = true;
            g_camera.setCurrentVelocity(velocity.x, velocity.y, 0.0f);
        }

        direction.z -= 1.0f;
    }
    else
    {
        moveBackwardsPressed = false;
    }

    if (keyboard.keyDown(Keyboard::KEY_RIGHT) || keyboard.keyDown(Keyboard::KEY_D))
    {
        if (!moveRightPressed)
        {
            moveRightPressed = true;
            g_camera.setCurrentVelocity(0.0f, velocity.y, velocity.z);
        }

        direction.x += 1.0f;
    }
    else
    {
        moveRightPressed = false;
    }

    if (keyboard.keyDown(Keyboard::KEY_LEFT) || keyboard.keyDown(Keyboard::KEY_A))
    {
        if (!moveLeftPressed)
        {
            moveLeftPressed = true;
            g_camera.setCurrentVelocity(0.0f, velocity.y, velocity.z);
        }

        direction.x -= 1.0f;
    }
    else
    {
        moveLeftPressed = false;
    }

	// Fly UP
    if (keyboard.keyDown(Keyboard::KEY_E) || keyboard.keyDown(Keyboard::KEY_PAGEUP))
    {
        if (!moveUpPressed)
        {
            moveUpPressed = true;
            g_camera.setCurrentVelocity(velocity.x, 0.0f, velocity.z);
        }

        direction.y += 1.0f;
    }
    else
    {
        moveUpPressed = false;
    }
	// Fly DOWN
    if (keyboard.keyDown(Keyboard::KEY_Q) || keyboard.keyDown(Keyboard::KEY_PAGEDOWN))
    {
        if (!moveDownPressed)
        {
            moveDownPressed = true;
            g_camera.setCurrentVelocity(velocity.x, 0.0f, velocity.z);
        }

        direction.y -= 1.0f;
    }
    else
    {
        moveDownPressed = false;
    }

    return direction;
}

bool Init()
{
    try
    {
        InitGL();

        if (!OpenGLSupportsGLVersion(2, 0))
            throw std::runtime_error("This application requires OpenGL 2.0 support.");

        InitApp();
        return true;
    }
    catch (const std::exception &e)
    {
        std::ostringstream msg;

        msg << "Application initialization failed!" << std::endl << std::endl;
        msg << e.what();

        MessageBox(g_hWnd, msg.str().c_str(), "Error", MB_ICONSTOP);
        return false;
    }
}

void InitApp()
{
	wireframe = false;
	drawWorld = true;
    
	if (!g_font.create("Courier New", 8,GLFont::BOLD))
        throw std::runtime_error("Failed to create font.");
    if (!(g_nullTexture = CreateNullTexture(2, 2)))
        throw std::runtime_error("failed to create null texture.");

    for (int i = 0; i < TERRAIN_REGIONS_COUNT; ++i)
    {
        if (!(g_regions[i].texture = LoadTexture(g_regions[i].filename.c_str())))
            throw std::runtime_error("Failed to load texture: " + g_regions[i].filename);
    }

    std::string infoLog;

    if (!(g_terrainShader = LoadShaderProgram("Data/shaders/terrain.glsl", infoLog)))
        throw std::runtime_error("Failed to load shader: terrain.glsl.\n" + infoLog);

    if (!g_terrain.create(HEIGHTMAP_SIZE, HEIGHTMAP_GRID_SPACING, HEIGHTMAP_SCALE))
        throw std::runtime_error("Failed to create terrain.");

	faultline = true; // start with faultline and low iteration to show pretty water
    GenerateTerrain(0);

    Vector3 pos;

	g_GodMode = true;
	// starts at a nice place in the air
    pos.x = HEIGHTMAP_SIZE * HEIGHTMAP_GRID_SPACING * 0.5f;
    pos.z = HEIGHTMAP_SIZE * HEIGHTMAP_GRID_SPACING ;//* 0.5f;   
	pos.y = 300;

    g_camera.setBehavior(Camera::CAMERA_BEHAVIOR_FIRST_PERSON);
    g_camera.setPosition(pos);
    g_camera.setAcceleration(CAMERA_ACCELERATION);
    g_camera.setVelocity(CAMERA_VELOCITY);

    g_camera.perspective(CAMERA_FOVX,
        static_cast<float>(g_windowWidth) / static_cast<float>(g_windowHeight),
        CAMERA_ZNEAR, CAMERA_ZFAR);

    float upperBounds = (HEIGHTMAP_SIZE * HEIGHTMAP_GRID_SPACING - (2.0f * HEIGHTMAP_GRID_SPACING));
    float lowerBounds = static_cast<float>(HEIGHTMAP_GRID_SPACING);

    g_cameraBoundsMax.x = upperBounds;
    g_cameraBoundsMax.y = CAMERA_ZFAR;
    g_cameraBoundsMax.z = upperBounds;

    g_cameraBoundsMin.x = lowerBounds;
    g_cameraBoundsMin.y = 0.0f;
    g_cameraBoundsMin.z = lowerBounds;

    Mouse::instance().hideCursor(true);
    Mouse::instance().setPosition(g_windowWidth / 2, g_windowHeight / 2);

	// Here we initialize our multitexturing functions
	glActiveTextureARB		= (PFNGLACTIVETEXTUREARBPROC)		wglGetProcAddress("glActiveTextureARB");
	glMultiTexCoord2fARB	= (PFNGLMULTITEXCOORD2FARBPROC)		wglGetProcAddress("glMultiTexCoord2fARB");

	CreateRenderTexture(g_Texture, g_TextureSize, 3, GL_RGB, REFLECTION_ID);
	CreateRenderTexture(g_Texture, g_TextureSize, 3, GL_RGB, REFRACTION_ID);
	CreateRenderTexture(g_Texture, g_TextureSize, 1, GL_DEPTH_COMPONENT, DEPTH_ID);

	CreateTexture(g_Texture[NORMAL_ID],	 "Data/textures//normalmap.bmp");// Load the normal map water texture
	CreateTexture(g_Texture[DUDVMAP_ID], "Data/textures//dudvmap.bmp");	// Load the dudv map water texture
	CreateTexture(g_Texture[TERRAIN_ID], "Data/textures//Terrain.bmp");	// Load the terrain texture
	CreateTexture(g_Texture[DETAIL_ID],	 "Data/textures//Detail.bmp");	// Load the detail texture for the terrain
	CreateTexture(g_Texture[BACK_ID],	 "Data/textures//skyboxback.bmp");		// Load the Sky box Back texture
	CreateTexture(g_Texture[FRONT_ID],	 "Data/textures//skyboxfront.bmp");	// Load the Sky box Front texture
	CreateTexture(g_Texture[BOTTOM_ID],  "Data/textures//skyboxbottom.bmp");	// Load the Sky box Bottom texture
	CreateTexture(g_Texture[TOP_ID],	 "Data/textures//skyboxtop.bmp");		// Load the Sky box Top texture
	CreateTexture(g_Texture[LEFT_ID],	 "Data/textures//skyboxleft.bmp");		// Load the Sky box Left texture
	CreateTexture(g_Texture[RIGHT_ID],	 "Data/textures//skyboxright.bmp");	// Load the Sky box Right texture

	LoadAnimTextures(g_Texture, "Data/textures//caust", WATER_START_ID, NUM_WATER_TEX);

	InitGLSL();

	g_Shader.InitShaders("Data/shaders//water.vert", "Data/shaders//water.frag");
	g_Shader.TurnOff();
	faultline = true; // start with faultline and low iteration to show pretty water
	g_displayHelp = true; // start with instructions
}

void InitGL()
{
    if (!(g_hDC = GetDC(g_hWnd)))
        throw std::runtime_error("GetDC() failed.");

    int pf = 0;
    PIXELFORMATDESCRIPTOR pfd = {0};
    OSVERSIONINFO osvi = {0};

    pfd.nSize = sizeof(pfd);
    pfd.nVersion = 1;
    pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
    pfd.iPixelType = PFD_TYPE_RGBA;
    pfd.cColorBits = 24;
    pfd.cDepthBits = 16;
    pfd.iLayerType = PFD_MAIN_PLANE;

    osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);

    if (!GetVersionEx(&osvi))
        throw std::runtime_error("GetVersionEx() failed."); 

    if (!pf)
        pf = ChoosePixelFormat(g_hDC, &pfd);

    if (!SetPixelFormat(g_hDC, pf, &pfd))
        throw std::runtime_error("SetPixelFormat() failed.");

    if (!(g_hRC = wglCreateContext(g_hDC)))
        throw std::runtime_error("wglCreateContext() failed.");

    if (!wglMakeCurrent(g_hDC, g_hRC))
        throw std::runtime_error("wglMakeCurrent() failed.");

    EnableVerticalSync(false);

    // Check for GL_EXT_texture_filter_anisotropic support.
    if (OpenGLExtensionSupported("GL_EXT_texture_filter_anisotropic"))
        glGetIntegerv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &g_maxAnisotrophy);
    else
        g_maxAnisotrophy = 1;
}

GLuint LinkShaders(GLuint vertShader, GLuint fragShader)
{
    GLuint program = glCreateProgram();

    if (program)
    {
        GLint linked = 0;

        if (vertShader)
            glAttachShader(program, vertShader);

        if (fragShader)
            glAttachShader(program, fragShader);

        glLinkProgram(program);
        glGetProgramiv(program, GL_LINK_STATUS, &linked);

        if (!linked)
        {
            GLsizei infoLogSize = 0;
            std::string infoLog;

            glGetProgramiv(program, GL_INFO_LOG_LENGTH, &infoLogSize);
            infoLog.resize(infoLogSize);
            glGetProgramInfoLog(program, infoLogSize, &infoLogSize, &infoLog[0]);

            throw infoLog;
        }
        if (vertShader)
            glDeleteShader(vertShader);

        if (fragShader)
            glDeleteShader(fragShader);
    }

    return program;
}

GLuint LoadShaderProgram(const char *pszFilename, std::string &infoLog)
{
    infoLog.clear();

    GLuint program = 0;
    std::string buffer;

    ReadTextFile(pszFilename, buffer);

    if (buffer.length() > 0)
    {
        const GLchar *pSource = 0;
        GLint length = 0;
        GLuint vertShader = 0;
        GLuint fragShader = 0;

        std::string::size_type vertOffset = buffer.find("[vert]");
        std::string::size_type fragOffset = buffer.find("[frag]");

        try
        {
            if (vertOffset != std::string::npos)
            {
                vertOffset += 6;        
                pSource = reinterpret_cast<const GLchar *>(&buffer[vertOffset]);
                length = static_cast<GLint>(fragOffset - vertOffset);
                vertShader = CompileShader(GL_VERTEX_SHADER, pSource, length);
            }
            if (fragOffset != std::string::npos)
            {
                fragOffset += 6;       
                pSource = reinterpret_cast<const GLchar *>(&buffer[fragOffset]);
                length = static_cast<GLint>(buffer.length() - fragOffset - 1);
                fragShader = CompileShader(GL_FRAGMENT_SHADER, pSource, length);
            }

             program = LinkShaders(vertShader, fragShader);
        }
        catch (const std::string &errors)
        {
            infoLog = errors;
        }
    }

    return program;
}

GLuint LoadTexture(const char *pszFilename)
{
    return LoadTexture(pszFilename, GL_LINEAR, GL_LINEAR_MIPMAP_LINEAR,
        GL_REPEAT, GL_REPEAT);
}

GLuint LoadTexture(const char *pszFilename, GLint magFilter, GLint minFilter,
                   GLint wrapS, GLint wrapT)
{
    GLuint id = 0;
    Bitmap bitmap;

    if (bitmap.loadPicture(pszFilename))
    {
        bitmap.flipVertical();

        glGenTextures(1, &id);
        glBindTexture(GL_TEXTURE_2D, id);

        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, magFilter);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, minFilter);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, wrapS);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, wrapT);
        glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE);

        if (g_maxAnisotrophy > 1)
        {
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT,
                g_maxAnisotrophy);
        }

        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, bitmap.width, bitmap.height,
            0, GL_BGRA, GL_UNSIGNED_BYTE, bitmap.getPixels());

        glBindTexture(GL_TEXTURE_2D, 0);
    }

    return id;
}

void PerformCameraCollisionDetection()
{
    const Vector3 &pos = g_camera.getPosition();
    Vector3 newPos(pos);

    if (pos.x > g_cameraBoundsMax.x)
        newPos.x = g_cameraBoundsMax.x;

    if (pos.x < g_cameraBoundsMin.x)
        newPos.x = g_cameraBoundsMin.x;

    if (pos.z > g_cameraBoundsMax.z)
        newPos.z = g_cameraBoundsMax.z;

    if (pos.z < g_cameraBoundsMin.z)
        newPos.z = g_cameraBoundsMin.z;

	// Don't want to go underground
	if(newPos.y < g_terrain.getHeightMap()->heightAt(newPos.x, newPos.z) + CAMERA_Y_OFFSET)
		newPos.y = g_terrain.getHeightMap()->heightAt(newPos.x, newPos.z) + CAMERA_Y_OFFSET; 

	// Don't want to go off too far 
	if(newPos.y > MAX_GOD_HEIGHT)		
		newPos.y = MAX_GOD_HEIGHT;	
	
	if(g_GodMode == false && newPos.y > g_terrain.getHeightMap()->heightAt(newPos.x, newPos.z) + CAMERA_Y_OFFSET + 1.5f) // to prevent wobbling
		newPos.y-=0.2f; // wicked feature to slowly drop camera when switching from god to no god mode

	else  if(!g_GodMode)
		newPos.y = g_terrain.getHeightMap()->heightAt(newPos.x, newPos.z) + CAMERA_Y_OFFSET;

    g_camera.setPosition(newPos);
}

void ProcessUserInput(float tick)
{
    Keyboard &keyboard = Keyboard::instance();

    if (keyboard.keyPressed(Keyboard::KEY_ESCAPE))
        PostMessage(g_hWnd, WM_CLOSE, 0, 0);

    if (keyboard.keyDown(Keyboard::KEY_LALT) || keyboard.keyDown(Keyboard::KEY_RALT))
    {
        if (keyboard.keyPressed(Keyboard::KEY_ENTER))
            ToggleFullScreen();
    }	
    if (keyboard.keyPressed(Keyboard::KEY_I))
        g_displayHelp = !g_displayHelp;

	// Increase/decrease iterations
    if (keyboard.keyDown(Keyboard::KEY_ADD) || keyboard.keyDown(Keyboard::KEY_NUMPAD_ADD))
    {
     	if(ITER>=MAX_ITER) ITER = MAX_ITER-1;
			else ITER++;
    }     
	if (keyboard.keyDown(Keyboard::KEY_SUBTRACT) || keyboard.keyDown(Keyboard::KEY_NUMPAD_SUBTRACT))
    {
      	if(ITER<=1) ITER = ITER+1;
			else ITER--;				
    }
		// Increase/decrease iterations
    if (keyboard.keyPressed(Keyboard::KEY_3))
    {
     	if(BLUR<=0) BLUR = 0;
			else BLUR-=0.05;
    }     
	if (keyboard.keyPressed(Keyboard::KEY_4))
    {
      	if(BLUR>=1) BLUR = 1;
			 else BLUR+=0.05;				
    }
    if (keyboard.keyPressed(Keyboard::KEY_V))
        EnableVerticalSync(!g_enableVerticalSync);

	if (keyboard.keyPressed(Keyboard::KEY_SPACE) || keyboard.keyPressed(Keyboard::KEY_N))
        GenerateTerrain(tick);	

    if (keyboard.keyPressed(Keyboard::KEY_T))
        //g_disableColorMaps = !g_disableColorMaps;
		drawWorld=!drawWorld;
	 if (keyboard.keyPressed(Keyboard::KEY_1))
       faultline = false;
	 if (keyboard.keyPressed(Keyboard::KEY_2))
       faultline = true;

	if (keyboard.keyPressed(Keyboard::KEY_G)) g_GodMode=!g_GodMode;
	if (keyboard.keyPressed(Keyboard::KEY_F)) 
	{
			wireframe = !wireframe;
		if (wireframe)
			glPolygonMode(GL_FRONT, GL_LINE);	
		else
			glPolygonMode(GL_FRONT, GL_FILL);
	}
}

void ReadTextFile(const char *pszFilename, std::string &buffer)
{
    std::ifstream file(pszFilename, std::ios::binary);

    if (file.is_open())
    {
        file.seekg(0, std::ios::end);

        std::ifstream::pos_type fileSize = file.tellg();

        buffer.resize(fileSize);
        file.seekg(0, std::ios::beg);
        file.read(&buffer[0], fileSize);
    }
}

void RenderFrame(float delta)
{
	CreateReflectionTexture(g_WaterHeight, g_TextureSize, delta);
	CreateRefractionDepthTexture(g_WaterHeight, g_TextureSize, delta);

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glViewport(0, 0, g_windowWidth, g_windowHeight);	
	glClearColor(0.3f, 0.5f, 0.9f, 0.0f); //nice colour for the water
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();	
	glMultMatrixf(&g_camera.getProjectionMatrix()[0][0]);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glMultMatrixf(&g_camera.getViewMatrix()[0][0]);
	g_Shader.TurnOn();
	glDisable(GL_CULL_FACE);
	RenderWater(g_WaterHeight);
	glEnable(GL_CULL_FACE);	
	g_Shader.TurnOff();
	if(drawWorld)RenderWorld(true, delta);
	RenderText();		
}

void RenderTerrain(float delta)
{
	glUseProgram(g_terrainShader);
    UpdateTerrainShaderParameters();

    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glLightfv(GL_LIGHT0, GL_POSITION, g_lightDir);
  
    if (g_disableColorMaps)
    {
        BindTexture(g_nullTexture, 0);
        BindTexture(g_nullTexture, 1);
        BindTexture(g_nullTexture, 2);
        BindTexture(g_nullTexture, 3);
    }
    else
    {
        BindTexture(g_regions[0].texture, 0);
        BindTexture(g_regions[1].texture, 1);
        BindTexture(g_regions[2].texture, 2);
        BindTexture(g_regions[3].texture, 3);
	}
    g_terrain.draw(delta);	    
    for (int i = 3; i >= 0; --i)
    {
        glActiveTexture(GL_TEXTURE0 + i);
        glBindTexture(GL_TEXTURE_2D, 0);
        glDisable(GL_TEXTURE_2D);
    }

    glDisable(GL_LIGHT0);
    glDisable(GL_LIGHTING);

    glUseProgram(0);
}

void RenderText()
{
std::ostringstream output;

    if (g_displayHelp)
    {
        output
            << "Press W and S to move forwards and backwards" << std::endl
            << "Press A and D to strafe left and right" << std::endl
			<< "Press Q and E to descend and descend in God Mode" << std::endl
            << "Move mouse to free look" << std::endl
            << std::endl
            << "Press 1 for Midpoint and 2 for Faultline algorithms" << std::endl
            << "Press T to enable/disable textures" << std::endl
            << "Press V to enable/disable vertical sync" << std::endl
			<< "Press F to enable/disable wireframe" << std::endl
            << "Press SPACE to generate a new random terrain" << std::endl
            << "Press +/- to change number of iterations for Fault line algorithm" << std::endl
			<< "Press 3/4 to decrease/increase blur amount for Fault line algorithm" << std::endl
            << "Press ALT + ENTER to toggle full screen" << std::endl
            << "Press ESC to exit" << std::endl
            << std::endl
            << "Press I to hide Instructions";
    }


    else
    {
        output.setf(std::ios::fixed, std::ios::floatfield);
        output << std::setprecision(2);

        output
            << "FPS: " << g_framesPerSecond << std::endl
			<< std::endl
       	    << "Current Algorithm: " << (returnFault() ? "Faultline" : "Midpoint") << std::endl
			<< "Iterations for Fault Line algorithm: " << ITER << std::endl
			<< "Blur amount (0-1) for Fault Line algorithm: " << BLUR << std::endl
			<< "God Mode:: " << (returnGod() ? "on" : "off") << std::endl
			<< "Vertical Sync: " << (returnVsync() ? "on" : "off") << std::endl
            << "Camera:" << std::endl
            << "  Position:"
            << " x:" << g_camera.getPosition().x
            << " y:" << g_camera.getPosition().y
            << " z:" << g_camera.getPosition().z << std::endl       
            << std::endl
            << "Press I for Instructions";
    }
    g_font.begin();
    g_font.setColor(0.0f, 1.0f, 0.0f);
    g_font.drawText(1, 1, output.str().c_str());
    g_font.end();
}

void SetProcessorAffinity()
{
    // Assign the current thread to one processor. This ensures that timing
    // code runs on only one processor, and will not suffer any ill effects
    // from power management.
    //
    // Based on DXUTSetProcessorAffinity() function from the DXUT framework.

    DWORD_PTR dwProcessAffinityMask = 0;
    DWORD_PTR dwSystemAffinityMask = 0;
    HANDLE hCurrentProcess = GetCurrentProcess();

    if (!GetProcessAffinityMask(hCurrentProcess, &dwProcessAffinityMask, &dwSystemAffinityMask))
        return;

    if (dwProcessAffinityMask)
    {
        // Find the lowest processor that our process is allowed to run against.

        DWORD_PTR dwAffinityMask = (dwProcessAffinityMask & ((~dwProcessAffinityMask) + 1));

        // Set this as the processor that our thread must always run against.
        // This must be a subset of the process affinity mask.

        HANDLE hCurrentThread = GetCurrentThread();

        if (hCurrentThread != INVALID_HANDLE_VALUE)
        {
            SetThreadAffinityMask(hCurrentThread, dwAffinityMask);
            CloseHandle(hCurrentThread);
        }
    }

    CloseHandle(hCurrentProcess);
}

void ToggleFullScreen()
{
    static DWORD savedExStyle;
    static DWORD savedStyle;
    static RECT rcSaved;

    g_isFullScreen = !g_isFullScreen;

    if (g_isFullScreen)
    {
        // Moving to full screen mode.

        savedExStyle = GetWindowLong(g_hWnd, GWL_EXSTYLE);
        savedStyle = GetWindowLong(g_hWnd, GWL_STYLE);
        GetWindowRect(g_hWnd, &rcSaved);

        SetWindowLong(g_hWnd, GWL_EXSTYLE, 0);
        SetWindowLong(g_hWnd, GWL_STYLE, WS_POPUP | WS_CLIPCHILDREN | WS_CLIPSIBLINGS);
        SetWindowPos(g_hWnd, HWND_TOPMOST, 0, 0, 0, 0,
            SWP_NOMOVE | SWP_NOSIZE | SWP_FRAMECHANGED | SWP_SHOWWINDOW);

        g_windowWidth = GetSystemMetrics(SM_CXSCREEN);
        g_windowHeight = GetSystemMetrics(SM_CYSCREEN);

        SetWindowPos(g_hWnd, HWND_TOPMOST, 0, 0,
            g_windowWidth, g_windowHeight, SWP_SHOWWINDOW);
    }
    else
    {
        // Moving back to windowed mode.

        SetWindowLong(g_hWnd, GWL_EXSTYLE, savedExStyle);
        SetWindowLong(g_hWnd, GWL_STYLE, savedStyle);
        SetWindowPos(g_hWnd, HWND_NOTOPMOST, 0, 0, 0, 0,
            SWP_NOMOVE | SWP_NOSIZE | SWP_FRAMECHANGED | SWP_SHOWWINDOW);

        g_windowWidth = rcSaved.right - rcSaved.left;
        g_windowHeight = rcSaved.bottom - rcSaved.top;

        SetWindowPos(g_hWnd, HWND_NOTOPMOST, rcSaved.left, rcSaved.top,
            g_windowWidth, g_windowHeight, SWP_SHOWWINDOW);
    }

    g_camera.perspective(CAMERA_FOVX,
        static_cast<float>(g_windowWidth) / static_cast<float>(g_windowHeight),
        CAMERA_ZNEAR, CAMERA_ZFAR);
}

void UpdateCamera(float elapsedTimeSec)
{
    Mouse &mouse = Mouse::instance();
    float dx = -mouse.xPosRelative();
    float dy = -mouse.yPosRelative();

    g_camera.rotateSmoothly(dx, dy, 0.0f);
    g_camera.updatePosition(GetMovementDirection(), elapsedTimeSec);

    mouse.setPosition(g_windowWidth / 2, g_windowHeight / 2);
	if(g_GodMode)
	{
		    g_camera.setAcceleration(CAMERA_GOD_ACCELERATION);
			g_camera.setVelocity(CAMERA_GOD_VELOCITY);
	}
	else
		
		    g_camera.setAcceleration(CAMERA_ACCELERATION);
			g_camera.setVelocity(CAMERA_VELOCITY);

	PerformCameraCollisionDetection();
}

void UpdateFrame(float elapsedTimeSec, float tick)
{
    UpdateFrameRate(elapsedTimeSec);

    Mouse::instance().update();
    Keyboard::instance().update();

    ProcessUserInput(tick);
    UpdateCamera(elapsedTimeSec);
}

void UpdateFrameRate(float elapsedTimeSec)
{
    static float accumTimeSec = 0.0f;
    static int frames = 0;

    accumTimeSec += elapsedTimeSec;

    if (accumTimeSec > 1.0f)
    {
        g_framesPerSecond = frames;

        frames = 0;
        accumTimeSec = 0.0f;
    }
    else
    {
        ++frames;
    }
}

void UpdateTerrainShaderParameters()
{
    GLint handle = -1;

    // Update the terrain tiling factor & regions
    handle = glGetUniformLocation(g_terrainShader, "tilingFactor");
    glUniform1f(handle, HEIGHTMAP_TILING_FACTOR);
    handle = glGetUniformLocation(g_terrainShader, "region1.max");
    glUniform1f(handle, g_regions[0].max);
    handle = glGetUniformLocation(g_terrainShader, "region1.min");
    glUniform1f(handle, g_regions[0].min);    
    handle = glGetUniformLocation(g_terrainShader, "region2.max");
    glUniform1f(handle, g_regions[1].max);
    handle = glGetUniformLocation(g_terrainShader, "region2.min");
    glUniform1f(handle, g_regions[1].min);
    handle = glGetUniformLocation(g_terrainShader, "region3.max");
    glUniform1f(handle, g_regions[2].max);
    handle = glGetUniformLocation(g_terrainShader, "region3.min");
    glUniform1f(handle, g_regions[2].min);
    handle = glGetUniformLocation(g_terrainShader, "region4.max");
    glUniform1f(handle, g_regions[3].max);
    handle = glGetUniformLocation(g_terrainShader, "region4.min");
    glUniform1f(handle, g_regions[3].min);

    // Bind textures.
    glUniform1i(glGetUniformLocation(g_terrainShader, "region1ColorMap"), 0);
    glUniform1i(glGetUniformLocation(g_terrainShader, "region2ColorMap"), 1);
    glUniform1i(glGetUniformLocation(g_terrainShader, "region3ColorMap"), 2);
    glUniform1i(glGetUniformLocation(g_terrainShader, "region4ColorMap"), 3);
}

bool CreateTexture(UINT &texture, LPSTR strFileName)
{
	AUX_RGBImageRec *pImage = NULL;
	FILE *pFile = NULL;

	if(!strFileName) 
		return false;
	if((pFile = fopen(strFileName, "rb")) == NULL) 
	{
		MessageBox(g_hWnd, "Unable to load BMP File!", "Error", MB_OK);
		return NULL;
	}
	pImage = auxDIBImageLoad(strFileName);				
	if(pImage == NULL)								
		return false;
	glGenTextures(1, &texture);
	glPixelStorei (GL_UNPACK_ALIGNMENT, 1);
	glBindTexture(GL_TEXTURE_2D, texture);
	gluBuild2DMipmaps(GL_TEXTURE_2D, 3, pImage->sizeX, 
					  pImage->sizeY, GL_RGB, GL_UNSIGNED_BYTE, pImage->data);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	if (pImage)										
	{
		if (pImage->data)							
		{
			free(pImage->data);						
		}

		free(pImage);								
	}
	return true;
}

void RenderWorld(bool bRenderCaustics, float delta)
{
	if(bRenderCaustics)
		RenderCaustics(g_WaterHeight, kCausticScale, delta);

	double TopTerrainPlane[4] = {0.0, 1.0, 0.0, -g_WaterHeight};
	// Turn a clipping plane on and assign our plane equation to it
	glEnable(GL_CLIP_PLANE0);
	glClipPlane(GL_CLIP_PLANE0, TopTerrainPlane);	
	// Render the clipped height map that is above the water
	RenderTerrain(delta);
	glDisable(GL_CLIP_PLANE0);	
	
	// Render our background sky box.
	CreateSkyBox(500, 0, 500, 3500, 3500, 3500);
	//CreateSkyBox(500+g_camera.getPosition().x,500+g_camera.getPosition().y, 500+g_camera.getPosition().z, 2500, 2500, 2500);
	
}

void CreateSkyBox(float x, float y, float z, float width, float height, float length)
{
	
	glEnable(GL_TEXTURE_2D);
	
	glBindTexture(GL_TEXTURE_2D, g_Texture[BACK_ID]);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

	x = x - width  / 2;
	y = y - height / 2;
	z = z - length / 2;


	glBegin(GL_QUADS);			
		//BACK
		glTexCoord2f(1.0f, 0.0f); glVertex3f(x + width, y,			z);
		glTexCoord2f(1.0f, 1.0f); glVertex3f(x + width, y + height, z); 
		glTexCoord2f(0.0f, 1.0f); glVertex3f(x,			y + height, z);
		glTexCoord2f(0.0f, 0.0f); glVertex3f(x,			y,			z);		
	glEnd();

	//FRONT
	glBindTexture(GL_TEXTURE_2D, g_Texture[FRONT_ID]);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glBegin(GL_QUADS);	
		glTexCoord2f(1.0f, 0.0f); glVertex3f(x,			y,			z + length);
		glTexCoord2f(1.0f, 1.0f); glVertex3f(x,			y + height, z + length);
		glTexCoord2f(0.0f, 1.0f); glVertex3f(x + width, y + height, z + length); 
		glTexCoord2f(0.0f, 0.0f); glVertex3f(x + width, y,			z + length);
	glEnd();

	//BOTTOM
	glBindTexture(GL_TEXTURE_2D, g_Texture[BOTTOM_ID]);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glBegin(GL_QUADS);	
		glTexCoord2f(1.0f, 0.0f); glVertex3f(x,			y,			z);
		glTexCoord2f(1.0f, 1.0f); glVertex3f(x,			y,			z + length);
		glTexCoord2f(0.0f, 1.0f); glVertex3f(x + width, y,			z + length); 
		glTexCoord2f(0.0f, 0.0f); glVertex3f(x + width, y,			z);
	glEnd();

	//TOP
	glBindTexture(GL_TEXTURE_2D, g_Texture[TOP_ID]);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glBegin(GL_QUADS);		
		glTexCoord2f(0.0f, 1.0f); glVertex3f(x + width, y + height, z);
		glTexCoord2f(0.0f, 0.0f); glVertex3f(x + width, y + height, z + length); 
		glTexCoord2f(1.0f, 0.0f); glVertex3f(x,			y + height,	z + length);
		glTexCoord2f(1.0f, 1.0f); glVertex3f(x,			y + height,	z);		
	glEnd();

	//LEFT
	glBindTexture(GL_TEXTURE_2D, g_Texture[LEFT_ID]);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glBegin(GL_QUADS);		
		glTexCoord2f(1.0f, 1.0f); glVertex3f(x,			y + height,	z);	
		glTexCoord2f(0.0f, 1.0f); glVertex3f(x,			y + height,	z + length); 
		glTexCoord2f(0.0f, 0.0f); glVertex3f(x,			y,			z + length);
		glTexCoord2f(1.0f, 0.0f); glVertex3f(x,			y,			z);				
	glEnd();

	//RIGHT
	glBindTexture(GL_TEXTURE_2D, g_Texture[RIGHT_ID]);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glBegin(GL_QUADS);		
		glTexCoord2f(0.0f, 0.0f); glVertex3f(x + width, y,			z);
		glTexCoord2f(1.0f, 0.0f); glVertex3f(x + width, y,			z + length);
		glTexCoord2f(1.0f, 1.0f); glVertex3f(x + width, y + height,	z + length); 
		glTexCoord2f(0.0f, 1.0f); glVertex3f(x + width, y + height,	z);
	glEnd();
}

